use crate::{gdev, Client};

use anyhow::{anyhow, Result};
use sp_core::{crypto::AccountId32, sr25519::Pair, Pair as _};
use std::ops::Deref;
use subxt::tx::{BaseExtrinsicParamsBuilder, PairSigner};

type SessionKeys = [u8; 128];

pub async fn rotate_keys(client: &Client) -> Result<SessionKeys> {
    client
        .rpc()
        .rotate_keys()
        .await?
        .deref()
        .try_into()
        .map_err(|e| anyhow!("Session keys have wrong length: {:?}", e))
}

pub async fn set_session_keys(pair: Pair, client: Client, session_keys: SessionKeys) -> Result<()> {
    client
        .tx()
        .sign_and_submit_then_watch(
            &gdev::tx()
                .authority_members()
                .set_session_keys(session_keys),
            &PairSigner::new(pair),
            BaseExtrinsicParamsBuilder::new(),
        )
        .await?;

    Ok(())
}

pub async fn update_session_keys(pair: Pair, client: Client) -> Result<()> {
    let session_keys = rotate_keys(&client).await?;
    set_session_keys(pair, client, session_keys).await
}

pub async fn go_online(pair: Pair, client: Client) -> Result<()> {
    if client
        .storage()
        .fetch(
            &gdev::storage()
                .session()
                .next_keys(AccountId32::from(pair.public())),
            None,
        )
        .await?
        .is_none()
    {
        return Err(anyhow!("This account has not set session keys!"));
    }

    client
        .tx()
        .sign_and_submit_then_watch(
            &gdev::tx().authority_members().go_online(),
            &PairSigner::new(pair),
            BaseExtrinsicParamsBuilder::new(),
        )
        .await?;

    Ok(())
}

pub async fn go_offline(pair: Pair, client: Client) -> Result<()> {
    client
        .tx()
        .sign_and_submit_then_watch(
            &gdev::tx().authority_members().go_offline(),
            &PairSigner::new(pair),
            BaseExtrinsicParamsBuilder::new(),
        )
        .await?;

    Ok(())
}
